#!/usr/bin/env python

"""
Amazon Fire TV server

RESTful interface for communication over a network via ADB
with Amazon Fire TV devices with ADB Debugging enabled.

From https://developer.amazon.com/public/solutions/devices/fire-tv/docs/connecting-adb-over-network:

Turn on ADB Debugging:
    1. From the main (Launcher) screen, select Settings.
    2. Select System > Developer Options.
    3. Select ADB Debugging.

Find device IP:
    1. From the main (Launcher) screen, select Settings.
    2. Select System > About > Network.
"""

import argparse
import re
import yaml
from flask import Flask, jsonify, request, abort
from firetv import FireTV


app = Flask(__name__)
devices = {}
config_data = None
valid_device_id = re.compile('^[-\w]+$')
valid_app_id = re.compile('^[A-Za-z0-9\.]+$')


def is_valid_host(host):
    """ Check if host is valid.

    Performs two simple checks:
        - Has host and port separated by ':'.
        - Port is a positive digit.

    :param host: Host in <address>:<port> format.
    :returns: Valid or not.
    """
    parts = host.split(':')
    return not (len(parts) != 2 or not parts[1].isdigit())


def is_valid_device_id(device_id):
    """ Check if device identifier is valid.

    A valid device identifier contains only ascii word characters or dashes.

    :param device_id: Device identifier
    :returns: Valid or not.
    """
    return valid_device_id.match(device_id)

def is_valid_app_id(app_id):
    """ check if app identifier is valid.

    To restrict access a valid app is one with only a-z, A-Z, and '.'.
    It is possible to make this less restrictive using the regex above.

    :param app_id: Application identifier
    :returns: Valid or not
    """
    return valid_app_id.match(app_id)

def add(device_id, host):
    """ Add a device.

    Creates FireTV instance associated with device identifier.

    :param device_id: Device identifier.
    :param host: Host in <address>:<port> format.
    :returns: Added successfully or not.
    """
    valid = is_valid_device_id(device_id) and is_valid_host(host)
    if valid:
        devices[device_id] = FireTV(str(host))
    return valid


@app.route('/devices/add', methods=['POST'])
def add_device():
    """ Add a device via HTTP POST.

    POST JSON in the following format ::

        {
            "device_id": "<your_device_id>",
            "host": "<address>:<port>"
        }

    """
    req = request.get_json()
    success = False
    if 'device_id' in req and 'host' in req:
        success = add(req['device_id'], req['host'])
    return jsonify(success=success)


@app.route('/devices/list', methods=['GET'])
def list_devices():
    """ List devices via HTTP GET. """
    output = {}
    for device_id, device in devices.items():
        output[device_id] = {
            'host': device.host,
            'state': device.state
        }
    return jsonify(devices=output)


@app.route('/devices/state/<device_id>', methods=['GET'])
def device_state(device_id):
    """ Get device state via HTTP GET. """
    if device_id not in devices:
        return jsonify(success=False)
    return jsonify(state=devices[device_id].state)

@app.route('/devices/<device_id>/apps/running', methods=['GET'])
def running_apps(device_id):
    """ Get running apps via HTTP GET. """
    if not is_valid_device_id(device_id):
        abort(403)
    if device_id not in devices:
        abort(404)
    return jsonify(running_apps=devices[device_id].running_apps())

@app.route('/devices/<device_id>/apps/state/<app_id>', methods=['GET'])
def get_app_state(device_id, app_id):
    """ Get the state of the requested app """
    if not is_valid_app_id(app_id):
        abort(403)
    if not is_valid_device_id(device_id):
        abort(403)
    if device_id not in devices:
        abort(404)
    app_state = devices[device_id].app_state(app_id)
    return jsonify(state=app_state, status=app_state)

@app.route('/devices/<device_id>/apps/<app_id>/state', methods=['GET'])
def get_app_state_alt(device_id, app_id):
    return get_app_state(device_id, app_id)

@app.route('/devices/action/<device_id>/<action_id>', methods=['GET'])
def device_action(device_id, action_id):
    """ Initiate device action via HTTP GET. """
    success = False
    if device_id in devices:
        input_cmd = getattr(devices[device_id], action_id, None)
        if callable(input_cmd):
            input_cmd()
            success = True
    return jsonify(success=success)

@app.route('/devices/<device_id>/apps/<app_id>/start' , methods=['GET'])
def app_start(device_id, app_id):
    """ Starts an app with corresponding package name"""
    if not is_valid_app_id(app_id):
        abort(403)
    if not is_valid_device_id(device_id):
        abort(403)
    if device_id not in devices:
        abort(404)
    if not devices[device_id].launch_app(app_id):
        abort(404)
    return jsonify(success=True)

@app.route('/devices/<device_id>/apps/<app_id>/stop' , methods=['GET'])
def app_stop(device_id, app_id):
    """ stops an app with corresponding package name"""
    if not is_valid_app_id(app_id):
        abort(403)
    if not is_valid_device_id(device_id):
        abort(403)
    if device_id not in devices:
        abort(404)
    devices[device_id].stop_app(app_id)
    return jsonify(success=True)

@app.route('/devices/connect/<device_id>', methods=['GET'])
def device_connect(device_id):
    """ Force a connection attempt via HTTP GET. """
    success = False
    if device_id in devices:
        devices[device_id].connect()
        success = True
    return jsonify(success=success)


def _parse_config(config_file_path):
    """ Parse Config File from yaml file. """
    config_file = open(config_file_path, 'r')
    config = yaml.load(config_file)
    config_file.close()
    return config

def _add_devices_from_config(args):
    """ Add devices from config. """
    config = _parse_config(args.config)
    for device in config['devices']:
        if args.default:
            if device == "default":
                raise ValueError('devicename "default" in config is not allowed if default param is set')
            if config['devices'][device]['host'] == args.default:
                raise ValueError('host set in default param must not be defined in config')
        add(device, config['devices'][device]['host'])

def main():
    """ Set up the server. """
    parser = argparse.ArgumentParser(description='AFTV Server')
    parser.add_argument('-p', '--port', type=int, help='listen port', default=5556)
    parser.add_argument('-d', '--default', help='default Amazon Fire TV host', nargs='?')
    parser.add_argument('-c', '--config', type=str, help='Path to config file')
    args = parser.parse_args()

    if args.config:
        _add_devices_from_config(args)

    if args.default and not add('default', args.default):
        exit('invalid hostname')
    app.run(host='0.0.0.0', port=args.port)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
Communicate with an Amazon Fire TV device via ADB over a network.

ADB Debugging must be enabled.
"""

import errno
from socket import error as socket_error
from adb import adb_commands
from adb.adb_protocol import InvalidChecksumError

# ADB key event codes.
HOME = 3
VOLUME_UP = 24
VOLUME_DOWN = 25
POWER = 26
PLAY_PAUSE = 85
NEXT = 87
PREVIOUS = 88
PLAY = 126
PAUSE = 127
UP = 19
DOWN = 20
LEFT = 21
RIGHT = 22
ENTER = 66
BACK = 4
MENU = 1

# Fire TV states.
STATE_ON = 'on'
STATE_IDLE = 'idle'
STATE_OFF = 'off'
STATE_PLAYING = 'play'
STATE_PAUSED = 'pause'
STATE_STANDBY = 'standby'
STATE_DISCONNECTED = 'disconnected'

class FireTV:
    """ Represents an Amazon Fire TV device. """

    def __init__(self, host):
        """ Initialize FireTV object.

        :param host: Host in format <address>:port.
        """
        self.host = host
        self._adb = None
        self.connect()

    def connect(self):
        """ Connect to an Amazon Fire TV device.

        Will attempt to establish ADB connection to the given host.
        Failure sets state to DISCONNECTED and disables sending actions.
        """
        try:
            self._adb = adb_commands.AdbCommands.ConnectDevice(
                serial=self.host)
        except socket_error as serr:
            if serr.errno != errno.ECONNREFUSED:
                raise serr

    @property
    def state(self):
        """ Compute and return the device state.

        :returns: Device state.
        """
        # Check if device is disconnected.
        if not self._adb:
            return STATE_DISCONNECTED
        # Check if device is off.
        if not self._screen_on:
            return STATE_OFF
        # Check if screen saver is on.
        if not self._awake:
            return STATE_IDLE
        # Check if the launcher is active.
        if self._launcher:
            return STATE_STANDBY
        # Check for a wake lock (device is playing).
        if self._wake_lock:
            return STATE_PLAYING
        # Otherwise, device is paused.
        return STATE_PAUSED

    def running_apps(self):
        """ Return an array of running user applications """
        return self._ps('u0_a')

    def app_state(self, app):
        """ Informs if application is running """
        if self.state == STATE_OFF or self.state == STATE_DISCONNECTED:
            return STATE_OFF
        if len(self._ps(app)) > 0:
            return STATE_ON
        return STATE_OFF

    def turn_on(self):
        """ Send power action if device is off. """
        if self.state == STATE_OFF:
            self._power()

    def turn_off(self):
        """ Send power action if device is not off. """
        if self.state != STATE_OFF:
            self._power()

    def home(self):
        """ Send home action. """
        self._key(HOME)

    def up(self):
        """ Send up action. """
        self._key(UP)

    def down(self):
        """ Send down action. """
        self._key(DOWN)

    def left(self):
        """ Send left action. """
        self._key(LEFT)

    def right(self):
        """ Send right action. """
        self._key(RIGHT)

    def enter(self):
        """ Send enter action. """
        self._key(ENTER)

    def back(self):
        """ Send back action. """
        self._key(BACK)

    def menu(self):
        """ Send menu action. """
        self._key(MENU)

    def volume_up(self):
        """ Send volume up action. """
        self._key(VOLUME_UP)

    def volume_down(self):
        """ Send volume down action. """
        self._key(VOLUME_DOWN)

    def media_play_pause(self):
        """ Send media play/pause action. """
        self._key(PLAY_PAUSE)

    def media_play(self):
        """ Send media play action. """
        self._key(PLAY)

    def media_pause(self):
        """ Send media pause action. """
        self._key(PAUSE)

    def media_next(self):
        """ Send media next action (results in fast-forward). """
        self._key(NEXT)

    def media_previous(self):
        """ Send media previous action (results in rewind). """
        self._key(PREVIOUS)

    def launch_app(self, app):
        if not self._adb:
            return
        mainActivity = self._mainActivity(app)
        if mainActivity:
            self._adb.Shell('am start -n {0}'.format(mainActivity))
            return True
        else:
            return False

    def stop_app(self, app):
        if not self._adb:
            return
        self._adb.Shell('am force-stop {0}'.format(app))

    @property
    def _screen_on(self):
        """ Check if the screen is on. """
        return self._dump_has('power', 'Display Power', 'state=ON')

    @property
    def _awake(self):
        """ Check if the device is awake (screen saver is not running). """
        return self._dump_has('power', 'mWakefulness', 'Awake')

    @property
    def _wake_lock(self):
        """ Check for wake locks (device is playing). """
        return not self._dump_has('power', 'Locks', 'size=0')

    @property
    def _launcher(self):
        """ Check if the active application is the Amazon TV launcher. """
        return self._dump_has('window', 'mFocusedApp=AppWindowToken',
                              'com.amazon.tv.launcher')



    def _power(self):
        """ Send power action. """
        self._key(POWER)

    def _input(self, cmd):
        """ Send input to the device.

        :param cmd: Input command.
        """
        if not self._adb:
            return
        self._adb.Shell('input {0}'.format(cmd))

    def _key(self, key):
        """ Send a key event to device.

        :param key: Key constant.
        """
        self._input('keyevent {0}'.format(key))

    def _dump(self, service, grep=None):
        """ Perform a service dump.

        :param service: Service to dump.
        :param grep: Grep for this string.
        :returns: Dump, optionally grepped.
        """
        if not self._adb:
            return
        if grep:
            return self._adb.Shell('dumpsys {0} | grep "{1}"'.format(service, grep))
        return self._adb.Shell('dumpsys {0}'.format(service))

    def _dump_has(self, service, grep, search):
        """ Check if a dump has particular content.

        :param service: Service to dump.
        :param grep: Grep for this string.
        :param search: Check for this substring.
        :returns: Found or not.
        """
        return self._dump(service, grep=grep).strip().find(search) > -1

    def _ps(self, search=''):
        """ Perform a ps command with optional filtering.

        :param search: Check for this substring.
        :returns: List of matching fields
        """
        if not self._adb:
            return
        result = []
        ps = self._adb.StreamingShell('ps')
        try:
            for bad_line in ps:
                # The splitting of the StreamingShell doesn't always work
                # this is to ensure that we get only one line
                for line in bad_line.splitlines():
                    if search in line:
                        result.append(line.strip().rsplit(' ',1)[-1])
            return result
        except InvalidChecksumError as e:
            print e
            self.connect()
            raise IOError

    def _mainActivity(self, app):
        return self._adb.Shell(('pm dump {0} | grep -A 1 "MAIN" | grep {0}').format(app)).strip().split(" ")[1]
